//structs shared between all effects
struct Light 
{
    float4 diffuse;
	float4 specular;
    float3 position;
    float3 direction;
    float falloff;
    float range;
    float attenuation0;
    float attenuation1;
    float attenuation2;
    float innerConeAngle;
    float outerConeAngle;
    int type; // 0: directional; 1: point; 2: spot;
    
};

struct VertexTextureShaderOutput
{
     float4 Position : POSITION;
     float2 TexCoords : TEXCOORD0;
     float3 WorldNormal : TEXCOORD1;
     float3 WorldPosition : TEXCOORD2;
	 float3 ViewDir : TEXCOORD3;
};

struct PixelShaderInput
{
     float2 TexCoords : TEXCOORD0;
     float3 WorldNormal : TEXCOORD1;
     float3 WorldPosition : TEXCOORD2;
	 float3 ViewDir : TEXCOORD3;
};

Light lights[12];
Light light;
int numberOfLights;

float2 scaleBias;

//shared scene parameters
shared float4x4 viewProjection;
shared float3 cameraPosition;
shared float4 ambientLightColor;

//texture parameters can be used to set states in the 
//effect state pass code
texture2D colorMapTexture;
texture2D normalMapTexture;
texture2D heightMapTexture;

sampler2D colorMapSampler = sampler_state
{
	Texture = <colorMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU  = wrap;
	AddressV  = wrap;
};

sampler2D normalMapSampler = sampler_state
{
    Texture = <normalMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU  = wrap;
	AddressV  = wrap;
};

sampler2D heightMapSampler = sampler_state
{
    Texture = <heightMapTexture>;
    MagFilter = Linear;
    MinFilter = Anisotropic;
    MipFilter = Linear;
    MaxAnisotropy = 16;
	AddressU  = wrap;
	AddressV  = wrap;
};

//the world paramter is not shared because it will
//change on every Draw() call
float4x4 world;
float4x4 worldForNormal;


//these material paramters are set per effect instance
float4 emissiveColor;
float4 diffuseColor;
float4 normalMapColor;
float4 specularColor;
float specularPower;
float specularIntensity;
bool colorMapEnabled;
bool normalMapEnabled;
bool heightMapEnabled;

static shared float4 compoundDiffuseColor;

// Alpha Blend Mixing
float DirectionalLightMix;
float SpotLightMix;
float PointLightMix;

float Transparency;

// Glow parameters
bool GlowEnabled;
float Inflate = 0.1;
float3 GlowColor = {1.0f, 0.9f, 0.3f};
float GlowExponential = 0.8;

//This function transforms the model to projection space and set up
//interpolators used by the pixel shader
VertexTextureShaderOutput BasicTextureVS(
     float3 position : POSITION,
     float3 normal : NORMAL,
     float2 texCoord : TEXCOORD0, 
     float3 binormal : BINORMAL0,
     float3 tangent   : TANGENT0)
{
     VertexTextureShaderOutput output;

     //generate the world-view-projection matrix
     float4x4 wvp = mul(world, viewProjection);
     
     //transform the input position to the output
     output.Position = mul(float4(position, 1.0), wvp);

     output.WorldNormal =  mul(normal, worldForNormal);
     output.WorldNormal = normalize(output.WorldNormal);
     float4 worldPosition =  mul(float4(position, 1.0), world);
     output.WorldPosition = worldPosition / worldPosition.w;
     
     //copy the tex coords to the interpolator
     output.TexCoords = texCoord;

	 float3 worldPos = mul(float4(position, 1.0f), world).xyz;
	 float3 viewDir = cameraPosition - worldPos;

	 float3x3 tbnMatrix = float3x3(tangent.x, binormal.x, normal.x,
	                               tangent.y, binormal.y, normal.y,
	                               tangent.z, binormal.z, normal.z);

	 output.ViewDir = mul(viewDir, tbnMatrix);
	 
     return output;
}

//The Ambient pixel shader simply adds an ambient color to the
//back buffer while outputting depth information.
float4 AmbientPS(PixelShaderInput input) : COLOR
{
	compoundDiffuseColor = diffuseColor;

	float3 v = normalize(input.ViewDir);

	if(colorMapEnabled)
    {
		float2 texCoord = input.TexCoords;
		float4 norm = float4(1, 1, 1, 1);

		if(heightMapEnabled)
		{
			float height = tex2D(heightMapSampler, input.TexCoords).r;
        
			height = height * scaleBias.x + scaleBias.y;
			texCoord = input.TexCoords + (height * v.xy);
		}

        compoundDiffuseColor *= tex2D(colorMapSampler, texCoord);		
    }
	
	float4 color = ambientLightColor * compoundDiffuseColor + emissiveColor;

	color.a = diffuseColor.a * (1 - Transparency);
		
	if(GlowEnabled)
	{
		float3 Nn = normalize(input.WorldNormal);
		float3 Vn = normalize(cameraPosition - input.WorldPosition);
		float edge = 1.0 - dot(Nn,Vn);
		edge = pow(abs(edge),GlowExponential);
		float3 result = edge + GlowColor.rgb;  

		return color * float4(result,edge);
	}  

	return color;
}


//This function calculates the diffuse and specular effect of a single light
//on a pixel given the world position, normal, and material properties
float4 CalculateSinglePointLight(Light light, float3 worldPosition, float3 worldNormal, 
                            float4 diffuseColor, float4 specularColor, float3 norm)
{    
	float3 lightVector = light.position - worldPosition;
	float lightDist = length(lightVector);
	float3 directionToLight = normalize(lightVector);
    
	float distanceAttenuation;                            
	distanceAttenuation = 1 / (light.attenuation0 + (light.attenuation1 + light.attenuation2 * lightDist) * lightDist);
		
     //calculate the intensity of the light with exponential falloff
     float baseIntensity = pow(saturate((light.range - lightDist) / light.range),
                                 light.falloff);
     
     baseIntensity *= distanceAttenuation;
     baseIntensity  = saturate(baseIntensity);
     
	 float nDotL = 1.0;

	 if(normalMapEnabled == true)
		nDotL = saturate(dot(norm, -directionToLight));

     float diffuseIntensity = saturate( dot(directionToLight, worldNormal));
     float4 diffuse = diffuseIntensity * light.diffuse * compoundDiffuseColor;

     //calculate Phong components per-pixel
     float3 reflectionVector = normalize(reflect(-directionToLight, worldNormal));
     float3 directionToCamera = normalize(cameraPosition - worldPosition);
     
     //calculate specular component
     float4 specular = saturate(light.specular  * specularColor *
                       pow(saturate(dot(reflectionVector, directionToCamera)), 
                           specularPower));
                           
     return baseIntensity * ((diffuse * nDotL) + specular);
}

float4 CalculateSingleDirectionalLight(Light light, float3 worldPosition, float3 worldNormal, 
									  float4 diffuseColor, float4 specularColor, float3 norm)
{
     float3 lightVector = -light.direction;
     float3 directionToLight = normalize(lightVector);
     
	 float nDotL = 1.0;

	 if(normalMapEnabled == true)
		nDotL = saturate(dot(norm, directionToLight));

     float diffuseIntensity = saturate( dot(directionToLight, worldNormal));
     float4 diffuse = diffuseIntensity * light.diffuse * compoundDiffuseColor;

     //calculate Phong components per-pixel
     float3 reflectionVector = normalize(reflect(-directionToLight, worldNormal));
     float3 directionToCamera = normalize(cameraPosition - worldPosition);
     
     //calculate specular component
     float4 specular = saturate(light.diffuse * light.specular * specularColor *
                       pow(saturate(dot(reflectionVector, directionToCamera)), 
                           specularPower));
                           
     return (diffuse * nDotL) + specular;
}

float4 CalculateSingleSpotLight(Light light, float3 worldPosition, float3 worldNormal, 
                            float4 diffuseColor, float4 specularColor, float3 norm)
{
    float3 lightVector = light.position - worldPosition;
	float lightDist = length(lightVector);
	float3 directionToLight = normalize(lightVector);

	float nDotL = 1.0;

	if(normalMapEnabled == true)
		nDotL = saturate(dot(norm, -light.direction));
    
	float distanceAttenuation;                            
	distanceAttenuation = 1 / (light.attenuation0 + (light.attenuation1 + light.attenuation2 * lightDist) * lightDist);
     
     float innerCos = cos(light.innerConeAngle / 2);
     float outerCos = cos(light.outerConeAngle / 2);
     float lightDirCos = dot(-directionToLight, normalize(light.direction));
     
     float coneAttenuation;
     if (lightDirCos > innerCos)
     {
		coneAttenuation = 1;
     }
     else if (lightDirCos < outerCos)
     {
		coneAttenuation = 0;
     }
     else
     {
		coneAttenuation = pow(abs((lightDirCos - outerCos) / (innerCos - outerCos)), light.falloff);
     }
     
     float diffuseIntensity = saturate( dot(directionToLight, worldNormal));
     float4 diffuse = diffuseIntensity * light.diffuse * compoundDiffuseColor;

     //calculate Phong components per-pixel
     float3 reflectionVector = normalize(reflect(-directionToLight, worldNormal));
     float3 directionToCamera = normalize(cameraPosition - worldPosition);
     
     //calculate specular component
     float4 specular = saturate(light.diffuse * specularColor *
                       pow(saturate(dot(reflectionVector, directionToCamera)), 
                           specularPower));
                           
	//return coneAttenuation;
     return  distanceAttenuation * coneAttenuation * ((diffuse * nDotL) + specular);
}


float4 MultipleDirectionalLightsPS(PixelShaderInput input) : COLOR
{
	compoundDiffuseColor = diffuseColor;

    float3 v = normalize(input.ViewDir);
	float3 norm = float3(1, 1, 1);

	if(colorMapEnabled)
    {
		float2 texCoord = input.TexCoords;
						

		if(heightMapEnabled)
		{
			float height = tex2D(heightMapSampler, input.TexCoords).r;
        
			height = height * scaleBias.x + scaleBias.y;
			texCoord = input.TexCoords + (height * v.xy);
		}

		if(normalMapEnabled)
		{
			norm = normalize(tex2D(normalMapSampler, texCoord).rgb * 2.0f - 1.0f);
		}

        compoundDiffuseColor *= tex2D(colorMapSampler, texCoord);		
    }
	
    float4 color = 0;
	int i = 0;
	
    for (; i< numberOfLights; i++)
    {    
		color += CalculateSingleDirectionalLight(lights[i], 
						 input.WorldPosition, input.WorldNormal,
						diffuseColor, specularColor, norm);
	}	

	color.a = DirectionalLightMix;

	if(GlowEnabled)
	{
		float3 Nn = normalize(input.WorldNormal);
		float3 Vn = normalize(cameraPosition - input.WorldPosition);
		float edge = 1.0 - dot(Nn,Vn);
		edge = pow(edge,GlowExponential);
		float3 result = edge * GlowColor.rgb;  

		return color + float4(result,edge);
	}  

	return color;	
}

float4 MultiplePointLightsPS(PixelShaderInput input) : COLOR
{
	compoundDiffuseColor = diffuseColor;

    float3 v = normalize(input.ViewDir);
	float3 norm = float3(1, 1, 1);

	if(colorMapEnabled)
    {
		float2 texCoord = input.TexCoords;
						

		if(heightMapEnabled)
		{
			float height = tex2D(heightMapSampler, input.TexCoords).r;
        
			height = height * scaleBias.x + scaleBias.y;
			texCoord = input.TexCoords + (height * v.xy);
		}

		if(normalMapEnabled)
		{
			norm = normalize(tex2D(normalMapSampler, texCoord).rgb * 2.0f - 1.0f);
		}

        compoundDiffuseColor *= tex2D(colorMapSampler, texCoord);		
    }
	
    float4 color = 0;
	int i = 0;
	
    for (; i< numberOfLights; i++)
    {    
		color += CalculateSinglePointLight(lights[i], 
						 input.WorldPosition, input.WorldNormal,
						diffuseColor, specularColor, norm);
	}
	
	color.a = PointLightMix;

	if(GlowEnabled)
	{
		float3 Nn = normalize(input.WorldNormal);
		float3 Vn = normalize(cameraPosition - input.WorldPosition);
		float edge = 1.0 - dot(Nn,Vn);
		edge = pow(edge,GlowExponential);
		float3 result = edge * GlowColor.rgb;  

		return color + float4(result,edge);
	}  

	return color;	
}

float4 MultipleSpotLightsPS(PixelShaderInput input) : COLOR
{
	compoundDiffuseColor = diffuseColor;
    
	float3 v = normalize(input.ViewDir);
	float3 norm = float3(1, 1, 1);

	if(colorMapEnabled)
    {
		float2 texCoord = input.TexCoords;
						

		if(heightMapEnabled)
		{
			float height = tex2D(heightMapSampler, input.TexCoords).r;
        
			height = height * scaleBias.x + scaleBias.y;
			texCoord = input.TexCoords + (height * v.xy);
		}

		if(normalMapEnabled)
		{
			norm = normalize(tex2D(normalMapSampler, texCoord).rgb * 2.0f - 1.0f);
		}

        compoundDiffuseColor *= tex2D(colorMapSampler, texCoord);		
    }
	
    float4 color = 0;
	int i = 0;
	
    for (; i< numberOfLights; i++)
    {    
		color += CalculateSingleSpotLight(lights[i], 
						 input.WorldPosition, input.WorldNormal,
						diffuseColor, specularColor, norm);
	}
	
	color.a = SpotLightMix;

	if(GlowEnabled)
	{
		float3 Nn = normalize(input.WorldNormal);
		float3 Vn = normalize(cameraPosition - input.WorldPosition);
		float edge = 1.0 - dot(Nn,Vn);
		edge = pow(abs(edge),GlowExponential);
		float3 result = edge * GlowColor.rgb;  

		return color + float4(result,edge);
	}  

	return color;	
}

technique GeneralTextureLighting
{
    pass Ambient
    {
        ////set sampler states
        //MagFilter[0] = LINEAR;
        //MinFilter[0] = LINEAR;
        //MipFilter[0] = LINEAR;
        //AddressU[0] = WRAP;
        //AddressV[0] = WRAP;
        //MagFilter[1] = LINEAR;
        //MinFilter[1] = LINEAR;
        //MipFilter[1] = LINEAR;
        //AddressU[1] = WRAP;
        //AddressV[1] = WRAP;
        //
        ////set texture states (notice the '<' , '>' brackets)
        ////as the texture state assigns a reference
        //Texture[0] = <colorMapTexture>;
       
        
        VertexShader = compile vs_2_0 BasicTextureVS();
        PixelShader = compile ps_2_0 AmbientPS();
    }

    pass MultipleDirectionalLight
    {
        VertexShader = compile vs_3_0 BasicTextureVS();
		PixelShader = compile ps_3_0 MultipleDirectionalLightsPS();
    }
    
    pass MultiplePointLight
    {
        VertexShader = compile vs_3_0 BasicTextureVS();
        PixelShader = compile ps_3_0 MultiplePointLightsPS();
    }
    pass MultipleSpotLight
    {
        VertexShader = compile vs_3_0 BasicTextureVS();
        PixelShader = compile ps_3_0 MultipleSpotLightsPS();
    }
}

