﻿/************************************************************************************ 
 * Copyright (c) 2008-2011, Columbia University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Columbia University nor the
 *       names of its contributors may be used to endorse or promote products
 *       derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY COLUMBIA UNIVERSITY ''AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * 
 * ===================================================================================
 * Author: Ohan Oda (ohan@cs.columbia.edu)
 * 
 *************************************************************************************/ 

using System;
using System.ComponentModel;
using System.Collections.Generic;
using System.Text;
using System.Xml;

using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using NOVA.Utilities;


namespace NOVA.Graphics
{
    /// <summary>
    /// An implementation of IModel interface for models created with CustomMesh.
    /// </summary>
    public class PrimitiveModel : IModel
    {
        #region Fields

        protected bool enabled;
        protected bool useLighting;
        protected bool castShadow;
        protected bool receiveShadow;
        protected bool useVertexColor;

        protected CustomMesh customMesh;
        protected BoundingBox boundingBox;
        protected BoundingSphere boundingSphere;
        protected bool showBoundingBox;
        protected Matrix offsetTransform;
        protected bool offsetToOrigin;
        protected int triangleCount;

        protected String technique;

        protected List<Vector3> vertices;
        protected List<int> indices;

        protected String shaderName;
        protected String customShapeParameters;

        #region Temporary Variables

        protected Matrix tmpMat1;
        protected Matrix tmpMat2;
        protected Vector3 tmpVec1;

        #endregion

        #endregion

        #region Constructors

        public PrimitiveModel() : this(null) { }

        /// <summary>
        /// Creates a model with VertexBuffer and IndexBuffer.
        /// </summary>
        /// <param name="customMesh">A mesh defined with VertexBuffer and IndexBuffer</param>
        public PrimitiveModel(CustomMesh customMesh)
        {
            offsetTransform = Matrix.Identity;
            offsetToOrigin = false;

            vertices = new List<Vector3>();
            indices = new List<int>();

            this.customMesh = customMesh;

            customShapeParameters = "";

            enabled = true;
            useLighting = true;
            castShadow = false;
            receiveShadow = false;
            showBoundingBox = false;
            useVertexColor = false;
            technique = "";

            CalculateMinimumBoundingBox();
            triangleCount = customMesh.NumberOfPrimitives;
        }

        #endregion

        #region Properties

        /// <summary>
        /// Flag indicating whether model is enabled and should be rendered. The default value is true.
        /// </summary>
        public bool Enabled
        {
            get { return enabled; }
            set { enabled = value; }
        }

        /// <summary>
        /// Flag reflecting whether lighting should be used when rendering this model. 
        /// The default value is true.
        /// </summary>
        public bool UseLighting
        {
            get { return useLighting; }
            set { useLighting = value; }
        }

        /// <summary>
        /// Gets or sets if this model can cast shadows on other objects that can receive shadows.
        /// The default value is false.
        /// </summary>
        public bool CastShadows
        {
            get { return castShadow; }
            set { castShadow = value; }
        }

        /// <summary>
        /// Gets or sets if this model can receive shadows cast by objects that can cast shadows.
        /// The default value is false.
        /// </summary>
        public bool ReceiveShadows
        {
            get { return receiveShadow; }
            set { receiveShadow = value; }
        }

        public String ShaderTechnique
        {
            get { return technique; }
            set { technique = value; }
        }

        public List<Vector3> Vertices
        {
            get { return vertices; }
        }

        public List<int> Indices
        {
            get { return indices; }
        }

        public PrimitiveType PrimitiveType
        {
            get{ return customMesh.PrimitiveType; }
        }

        /// <summary>
        /// Gets the mesh defined with VertexBuffer and IndexBuffer .
        /// </summary>
        public CustomMesh CustomMesh
        {
            get { return customMesh; }
        }

        /// <summary>
        /// Gets the minimum bounding box used for display and by the physics engine.
        /// </summary>
        public BoundingBox MinimumBoundingBox
        {
            get { return boundingBox; }
        }

        /// <summary>
        /// Gets the minimum bounding sphere that bounds this model mesh
        /// </summary>
        public BoundingSphere MinimumBoundingSphere
        {
            get { return boundingSphere; }
        }

        /// <summary>
        /// Gets the offset transformation from the origin of the world coordinate.
        /// </summary>
        /// <remarks>
        /// If not provided, this transform will be calculated based on the following equation:
        /// OffsetTranslation.Translation = (MinimumBoundingBox.Min + MinimumBoundingBox.Max) / 2.
        /// In this case, no rotation offset will be calculated.
        /// </remarks>
        public Matrix OffsetTransform
        {
            get { return offsetTransform; }
            internal set { offsetTransform = value; }
        }
        
        /// <summary>
        /// Gets or sets whether to relocate the model to the origin. Each model has its 
        /// position stored in the model file, but if you want to relocate the model to the 
        /// origin instead of locating it based on the position stored in the file, you should
        /// set this to true. The default value is false.
        /// </summary>
        public bool OffsetToOrigin
        {
            get { return offsetToOrigin; }
            set { offsetToOrigin = value; }
        }

        /// <summary>
        /// Gets or sets whether to draw the minimum bounding box around the model.
        /// The default value is false.
        /// </summary>
        public bool ShowBoundingBox
        {
            get { return showBoundingBox; }
            set { showBoundingBox = value; }
        }

        /// <summary>
        /// Gets the triangle count of this model
        /// </summary>
        public int TriangleCount
        {
            get { return triangleCount; }
        }

        /// <summary>
        /// Gets or sets the name of the shader used to illuminate this model. The default value 
        /// is SimpleEffectShader.
        /// </summary>
        /// <remarks>
        /// This information is necessary for saving and loading scene graph from an XML file.
        /// </remarks>
        public String ShaderName
        {
            get { return shaderName; }
            set { shaderName = value; }
        }

        /// <summary>
        /// Gets or sets the parameters needed to be passed to a class that contructs a primitive
        /// shape. 
        /// </summary>
        /// <remarks>
        /// This information is necessary for saving and loading scene graph from an XML file.
        /// </remarks>
        public String CustomShapeParameters
        {
            get { return customShapeParameters; }
            set { customShapeParameters = value; }
        }

        #endregion

        #region Public Methods

        /// <summary>
        /// Calculates the minimum bounding box that fits this model.
        /// </summary>
        protected virtual void CalculateMinimumBoundingBox()
        {
            int stride = customMesh.VertexDeclaration.VertexStride;
            int numberv = customMesh.NumberOfVertices;
            byte[] data = new byte[stride * numberv];

            customMesh.VertexBuffer.GetData<byte>(data);

            for (int ndx = 0; ndx < data.Length; ndx += stride)
            {
                tmpVec1.X = BitConverter.ToSingle(data, ndx);
                tmpVec1.Y = BitConverter.ToSingle(data, ndx + 4);
                tmpVec1.Z = BitConverter.ToSingle(data, ndx + 8);
                vertices.Add(tmpVec1);
            }

            if (customMesh.IndexBuffer.BufferUsage == BufferUsage.None)
            {
                short[] tmpIndices = new short[customMesh.IndexBuffer.IndexCount];
                customMesh.IndexBuffer.GetData<short>(tmpIndices);
                int[] tmpIntIndices = new int[tmpIndices.Length];
                Array.Copy(tmpIndices, tmpIntIndices, tmpIndices.Length);
                indices.AddRange(tmpIntIndices);
            }

            if (vertices.Count == 0)
            {
                ExceptionManager.Add(new Exception("Corrupted model vertices. Failed to calculate MBB."));
            }
            else
            {
                boundingBox = BoundingBox.CreateFromPoints(vertices);
                boundingSphere = BoundingSphere.CreateFromPoints(vertices);
                if(offsetTransform.Equals(Matrix.Identity))
                    offsetTransform.Translation = -(boundingBox.Min + boundingBox.Max) / 2;
            }
        }

        protected virtual void RenderBoundingBox(Matrix renderMatrix)
        {
            Vector3[] corners = boundingBox.GetCorners();
            for (int i = 0; i < corners.Length; i++)
                Vector3.Transform(ref corners[i], ref renderMatrix, out corners[i]);

            DebugShapeRenderer.AddBoundingBox(corners, Core.BoundingBoxColor, 0);
        }

        /// <summary>
        /// Disposes of model contents.
        /// </summary>
        public virtual void Dispose()
        {
            vertices.Clear();
        }

        /// <summary>
        /// Renders all objects that should receive shadows here. Called from the 
        /// ShadowMappingShader.UseShadow method.
        /// </summary>
        /// <remarks>
        /// This function is called automatically to receive shadows, so do not call this method
        /// </remarks>
        /// <param name="renderMatrix">Transform of this model</param>
        public virtual void UseShadows(Matrix renderMatrix)
        {
            // The model only receives shadow if receiveShadow is set to true
            if (!receiveShadow)
                return;

            /*State.ShadowShader.UpdateCalcShadowWorldMatrix(renderMatrix);
            State.Device.Vertices[0].SetSource(
                customMesh.VertexBuffer, 0, customMesh.SizeInBytes);
            State.Device.Indices = customMesh.IndexBuffer;
            State.Device.VertexDeclaration = customMesh.VertexDeclaration;
            State.Device.DrawIndexedPrimitives(customMesh.PrimitiveType,
                0, 0, customMesh.NumberOfVertices, 0, customMesh.NumberOfPrimitives);*/
        }

        /// <summary>
        /// Generate shadows for this model in the generate shadows pass
        /// of our shadow mapping shader. All objects rendered here will
        /// cast shadows in our scene.
        /// </summary>
        /// <remarks>
        /// This function is called automatically to cast shadows, so do not call this method
        /// </remarks>
        /// <param name="renderMatrix">Transform of this model</param>
        public virtual void GenerateShadows(Matrix renderMatrix)
        {
            if (!castShadow)
                return;

            /*State.ShadowShader.UpdateGenerateShadowWorldMatrix(renderMatrix);
            State.Device.Vertices[0].SetSource(
                customMesh.VertexBuffer, 0, customMesh.SizeInBytes);
            State.Device.Indices = customMesh.IndexBuffer;
            State.Device.VertexDeclaration = customMesh.VertexDeclaration;
            State.Device.DrawIndexedPrimitives(customMesh.PrimitiveType,
                0, 0, customMesh.NumberOfVertices, 0, customMesh.NumberOfPrimitives);*/
        }

        #endregion
    }
}
